/*samplerDriver.c */
/* samplerDriver.c,v 1.2 2005/11/18 16:16:21 xusf Exp */
/***********************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of
* California, as Operator of Los Alamos National Laboratory, and
* Berliner Elektronenspeicherring-Gesellschaft m.b.H. (BESSY).
* asynDriver is distributed subject to a Software License Agreement
* found in file LICENSE that is included with this distribution.
***********************************************************************/
/* driver support
 * Author: Shifu Xu
 * Date:   03NOV2005
 */

#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <math.h>

#include "epicsPrint.h"

#include <ctype.h>
#include <cantProceed.h>
#include <epicsStdio.h>
#include <epicsMutex.h>
#include <epicsEvent.h>
#include <epicsThread.h>
#include <epicsExport.h>
#include <iocsh.h>
#include <asynDriver.h>
#include <asynDrvUser.h>
#include <asynInt32.h>
#include <asynInt32Array.h>
#include "sampler.h"
#define NCHANNELS 37

typedef enum requestType
{
    TOTAL_SAMPLES,
    MAX_INTERRUPT_LATENCY,
    MAX_CONTEXT_SWITCH,
    NUM_GTM_INTERRUPT_LATENCY,
    NUM_GTM_CONTEXT_SWITCH,
    SAMPLES,
    START,
    CLEAR,
    INTERRUPT_LATENCY,
    CONTEXT_SWITCH,
    CUTOFF_H,
    CUTOFF_V,
    CUTOFF_H1,
    CUTOFF_V1,

    CUTOFF_H_TOP_TO_BH,
    CUTOFF_V_TOP_TO_BH,
    CUTOFF_H_BH_TO_USER,
    CUTOFF_V_BH_TO_USER,
    CUTOFF_H_TOTAL_RESPONSE,
    CUTOFF_V_TOTAL_RESPONSE,

    MAX_TOP_TO_BH,
    MAX_BH_TO_USER,
    NUM_GTM_TOP_TO_BH,
    NUM_GTM_BH_TO_USER,
    TOP_TO_BH,
    BH_TO_USER,
    MAX_TOTAL_RESPONSE,
    NUM_GTM_TOTAL_RESPONSE,
    TOTAL_RESPONSE,

    MIN_INTERRUPT_LATENCY,
    MIN_CONTEXT_SWITCH,
    MIN_TOP_TO_BH,
    MIN_BH_TO_USER,
    MIN_TOTAL_RESPONSE,

    MEDIAN_INTERRUPT_LATENCY,
    MEDIAN_CONTEXT_SWITCH,
    MEDIAN_TOP_TO_BH,
    MEDIAN_BH_TO_USER,
    MEDIAN_TOTAL_RESPONSE
} requestType;

typedef struct mapType
{
    requestType type;
    char *requestMsg;
} mapType;

static mapType map[] = {
    {TOTAL_SAMPLES, "total_samples"},
    {MAX_INTERRUPT_LATENCY, "max_interrupt_latency"},
    {MAX_CONTEXT_SWITCH, "max_context_switch"},
    {NUM_GTM_INTERRUPT_LATENCY, "num_gtm_interrupt_latency"},
    {NUM_GTM_CONTEXT_SWITCH, "num_gtm_context_switch"},
    {SAMPLES, "samples"},
    {START, "start"},
    {CLEAR, "clear"},
    {INTERRUPT_LATENCY, "interrupt_latency"},
    {CONTEXT_SWITCH, "context_switch"},

    {CUTOFF_H, "cutoff_h"},
    {CUTOFF_V, "cutoff_v"},
    {CUTOFF_H1, "cutoff_h1"},
    {CUTOFF_V1, "cutoff_v1"},
    {CUTOFF_H_TOP_TO_BH, "cutoff_h_top_to_bh"},
    {CUTOFF_V_TOP_TO_BH, "cutoff_v_top_to_bh"},
    {CUTOFF_H_BH_TO_USER, "cutoff_h_bh_to_user"},
    {CUTOFF_V_BH_TO_USER, "cutoff_v_bh_to_user"},
    {CUTOFF_H_TOTAL_RESPONSE, "cutoff_h_total_response"},
    {CUTOFF_V_TOTAL_RESPONSE, "cutoff_v_total_response"},

    {MAX_TOP_TO_BH, "max_top_to_bh"},
    {MAX_BH_TO_USER, "max_bh_to_user"},
    {NUM_GTM_TOP_TO_BH, "num_gtm_top_to_bh"},
    {NUM_GTM_BH_TO_USER, "num_gtm_bh_to_user"},
    {TOP_TO_BH, "top_to_bh"},
    {BH_TO_USER, "bh_to_user"},
    {MAX_TOTAL_RESPONSE, "max_total_response"},
    {NUM_GTM_TOTAL_RESPONSE, "num_gtm_total_response"},
    {TOTAL_RESPONSE, "total_response"},

    {MIN_INTERRUPT_LATENCY, "min_interrupt_latency"},
    {MIN_CONTEXT_SWITCH, "min_context_switch"},
    {MIN_TOP_TO_BH, "min_top_to_bh"},
    {MIN_BH_TO_USER, "min_bh_to_user"},
    {MIN_TOTAL_RESPONSE, "min_total_response"},

    {MEDIAN_INTERRUPT_LATENCY, "median_interrupt_latency"},
    {MEDIAN_CONTEXT_SWITCH, "median_context_switch"},
    {MEDIAN_TOP_TO_BH, "median_top_to_bh"},
    {MEDIAN_BH_TO_USER, "median_bh_to_user"},
    {MEDIAN_TOTAL_RESPONSE, "median_total_response"},

};

typedef struct delayData
{
    unsigned long max;
    unsigned long min;
    unsigned long median;
    unsigned long numGtLimit;
    unsigned long cutoffH;
    unsigned long cutoffV;
    unsigned long arraySize;
    unsigned long *array;
} delayData;

typedef struct samplerPvt
{
    unsigned long totalSamples;
    unsigned long samples;
    epicsMutexId lock;
    delayData interruptLatency;
    delayData topToBottom;
    delayData bottomToUserInterrupt;
    delayData contextSwitch;
    delayData totalResponse;

    char samplerName[20];
    getSampleType getSample;
    void *samplerPvt;
    delayMeasure onceMeasure;
} samplerPvt;
static samplerPvt *psamplerPvt;

typedef struct drvPvt
{
    const char *portName;
    int connected;
    asynInterface common;
    asynInterface asynDrvUser;
    asynInterface asynInt32;
    asynInterface asynInt32Array;
    void *asynInt32Pvt;		/* For registerInterruptSource */
    void *asynInt32ArrayPvt;	/* For registerInterruptSource */
} drvPvt_t;

typedef struct drvUserData
{
    requestType type;
} drvUserData;

static asynStatus getAddr (drvPvt_t * pdrvPvt, asynUser * pasynUser,
			   int *paddr, int portOK);
/* asynCommon methods */
static void report (void *drvPvt, FILE * fp, int details);
static asynStatus connect (void *drvPvt, asynUser * pasynUser);
static asynStatus disconnect (void *drvPvt, asynUser * pasynUser);
static asynCommon common = { report, connect, disconnect };

/* asynInt32 methods */
static asynStatus int32Write (void *drvPvt, asynUser * pasynUser,
			      epicsInt32 value);
static asynStatus int32Read (void *drvPvt, asynUser * pasynUser,
			     epicsInt32 * value);
/*use default for registerCallback,cancelCallback*/
static asynInt32 int32 = { int32Write, int32Read, 0, 0, 0 };

/* asynInt32Array methods */
static asynStatus int32ArrayRead (void *drvPvt, asynUser * pasynUser,
				  epicsInt32 * value, size_t nelements,
				  size_t * nIn);
/*use default for registerCallback,cancelCallback*/
static asynInt32Array int32Array = { 0, int32ArrayRead, 0, 0 };

int samplerDriverInit (const char *dn, int arraySizeInterruptLatency,
		       int arraySizeTopTobottom,
		       int arraySizeBottomToUserInterrupt,
		       int arraySizeContextSwitch, int arraySizeTotalResponse);
static int registerSampler (const char *name, getSampleType sampler,
			    void *samplerPvt);
static int removeSampler (void *samplerPvt);
static asynStatus create (void *drvPvt, asynUser * pasynUser,
			  const char *drvInfo, const char **pptypeName,
			  size_t * psize);
static asynStatus getType (void *ddrvPvt, asynUser * pasynUser,
			   const char **pptypeName, size_t * psize);
static asynStatus destroy (void *drvPvt, asynUser * pasynUser);

static asynDrvUser drvUser = { create, getType, destroy };

static samplerManager samplerManagerBase = {
    &registerSampler,
    &removeSampler,
};
samplerManager *psamplerManager = &samplerManagerBase;

static int registerSampler (const char *name, getSampleType getSample,
			    void *samplerPvt)
{
    strcpy (psamplerPvt->samplerName, name);
    psamplerPvt->getSample = getSample;
    psamplerPvt->samplerPvt = samplerPvt;
    return 0;
}

static int removeSampler (void *samplerPvt)
{
    strcpy (psamplerPvt->samplerName, "");
    psamplerPvt->getSample = NULL;
    psamplerPvt->samplerPvt = samplerPvt;
    return 0;
}

static asynStatus create (void *drvPvt, asynUser * pasynUser,
			  const char *drvInfo, const char **pptypeName,
			  size_t * psize)
{
    int i, j;
    drvUserData *pdrvUserData;
    pdrvUserData =
	callocMustSucceed (sizeof (drvUserData) + 1, sizeof (char),
			   "drvUserData create");

    j = sizeof (map) / sizeof (mapType);
    for (i = 0; i < j; i++)
      {
	  if (strcmp (map[i].requestMsg, drvInfo) == 0)
	      break;
      }
    if (i < j)
      {
	  pdrvUserData->type = map[i].type;
	  pasynUser->drvUser = (void *) pdrvUserData;
	  return asynSuccess;
      }
    else
      {
	  printf ("asyn userData,create(): no such type %s\n", drvInfo);
	  return asynError;
      }
}

static asynStatus getType (void *ddrvPvt, asynUser * pasynUser,
			   const char **pptypeName, size_t * psize)
{
    return asynSuccess;

}
static asynStatus destroy (void *drvPvt, asynUser * pasynUser)
{
    return asynSuccess;

}

int samplerDriverInit (const char *dn, int arraySizeInterruptLatency,
		       int arraySizeTopTobottom,
		       int arraySizeBottomToUserInterrupt,
		       int arraySizeContextSwitch, int arraySizeTotalResponse)
{
    drvPvt_t *pdrvPvt;
    char *portName;
    asynStatus status;
    int nbytes;

    nbytes = sizeof (samplerPvt) + 1;
    psamplerPvt =
	callocMustSucceed (nbytes, sizeof (char), "samplerDriverInit");

    psamplerPvt->samples = 1000;

    psamplerPvt->interruptLatency.cutoffH = 0;
    psamplerPvt->interruptLatency.cutoffV = 0;
    psamplerPvt->contextSwitch.cutoffH = 0;
    psamplerPvt->contextSwitch.cutoffV = 0;
    psamplerPvt->topToBottom.cutoffH = 0;
    psamplerPvt->topToBottom.cutoffV = 0;
    psamplerPvt->bottomToUserInterrupt.cutoffH = 0;
    psamplerPvt->bottomToUserInterrupt.cutoffV = 0;
    psamplerPvt->totalResponse.cutoffH = 0;
    psamplerPvt->totalResponse.cutoffV = 0;

    psamplerPvt->lock = epicsMutexMustCreate ();

    psamplerPvt->interruptLatency.arraySize = arraySizeInterruptLatency;
    psamplerPvt->topToBottom.arraySize = arraySizeTopTobottom;
    psamplerPvt->bottomToUserInterrupt.arraySize =
	arraySizeBottomToUserInterrupt;
    psamplerPvt->contextSwitch.arraySize = arraySizeContextSwitch;
    psamplerPvt->totalResponse.arraySize = arraySizeTotalResponse;

    psamplerPvt->interruptLatency.array =
	(unsigned long *) calloc (arraySizeInterruptLatency,
				  sizeof (unsigned long));
    psamplerPvt->topToBottom.array =
	(unsigned long *) calloc (arraySizeTopTobottom, sizeof (unsigned long));
    psamplerPvt->bottomToUserInterrupt.array =
	(unsigned long *) calloc (arraySizeBottomToUserInterrupt,
				  sizeof (unsigned long));
    psamplerPvt->contextSwitch.array =
	(unsigned long *) calloc (arraySizeContextSwitch,
				  sizeof (unsigned long));
    psamplerPvt->totalResponse.array =
	(unsigned long *) calloc (arraySizeTotalResponse,
				  sizeof (unsigned long));

    if (!psamplerPvt->interruptLatency.array || !psamplerPvt->topToBottom.array
	|| !psamplerPvt->bottomToUserInterrupt.array
	|| !psamplerPvt->contextSwitch.array
	|| !psamplerPvt->totalResponse.array)
      {

	  epicsPrintf ("samplerPvt: locate memory  failed\n");
/*	exit(1); */
      }

    nbytes = sizeof (drvPvt_t) + strlen (dn) + 1;
    pdrvPvt =
	callocMustSucceed (nbytes, sizeof (char), "samplerAsynDriverInit");
    portName = (char *) (pdrvPvt + 1);
    strcpy (portName, dn);
    pdrvPvt->portName = portName;
    pdrvPvt->common.interfaceType = asynCommonType;
    pdrvPvt->common.pinterface = (void *) &common;
    pdrvPvt->common.drvPvt = pdrvPvt;
    pdrvPvt->asynDrvUser.interfaceType = asynDrvUserType;
    pdrvPvt->asynDrvUser.pinterface = (void *) &drvUser;
    pdrvPvt->asynDrvUser.drvPvt = pdrvPvt;

    pdrvPvt->asynInt32.interfaceType = asynInt32Type;
    pdrvPvt->asynInt32Array.interfaceType = asynInt32ArrayType;
    pdrvPvt->asynInt32Array.pinterface = (void *) &int32Array;
    pdrvPvt->asynInt32Array.drvPvt = pdrvPvt;
    pdrvPvt->asynInt32.pinterface = (void *) &int32;
    pdrvPvt->asynInt32.drvPvt = pdrvPvt;
    status =
	pasynManager->registerPort (portName, ASYN_MULTIDEVICE | ASYN_CANBLOCK,
				    1, 100, 0);
    if (status != asynSuccess)
      {
	  printf ("samplerDriverInit registerDriver failed\n");
	  return 0;
      }
    status = pasynManager->registerInterface (portName, &pdrvPvt->common);
    if (status != asynSuccess)
      {
	  printf ("samplerDriverInit registerInterface failed\n");
	  return 0;
      }
    status = pasynManager->registerInterface (portName, &pdrvPvt->asynDrvUser);
    if (status != asynSuccess)
      {
	  printf ("samplerDriverInit registerInterface failed\n");
	  return 0;
      }
    status =
	pasynInt32Base->initialize (pdrvPvt->portName, &pdrvPvt->asynInt32);
    if (status != asynSuccess)
	return 0;
    status = pasynInt32ArrayBase->initialize (pdrvPvt->portName,
					      &pdrvPvt->asynInt32Array);
    if (status != asynSuccess)
	return 0;
    status =
	pasynManager->registerInterruptSource (portName, &pdrvPvt->asynInt32,
					       &pdrvPvt->asynInt32Pvt);
    if (status != asynSuccess)
      {
	  printf ("samplerDriverInit registerInterruptSource failed\n");
      }
    return (0);
}

static asynStatus getAddr (drvPvt_t * pdrvPvt, asynUser * pasynUser,
			   int *paddr, int portOK)
{
    asynStatus status;
    status = pasynManager->getAddr (pasynUser, paddr);
    if (status != asynSuccess)
	return status;
    if (*paddr >= -1 && *paddr < NCHANNELS)
	return asynSuccess;
    if (!portOK && *paddr >= 0)
	return asynSuccess;
    epicsSnprintf (pasynUser->errorMessage, pasynUser->errorMessageSize,
		   "%s addr %d is illegal; Must be >= %d and < %d\n",
		   pdrvPvt->portName, *paddr, (portOK ? -1 : 0), NCHANNELS);
    return asynError;
}

/* asynCommon methods */
static void report (void *pvt, FILE * fp, int details)
{
    drvPvt_t *pdrvPvt = (drvPvt_t *) pvt;
    fprintf (fp, "    samplerDriver: connected:%s \n",
	     (pdrvPvt->connected ? "Yes" : "No"));
}
static asynStatus connect (void *pvt, asynUser * pasynUser)
{
    drvPvt_t *pdrvPvt = (drvPvt_t *) pvt;
    int addr;
    asynStatus status;
    status = getAddr (pdrvPvt, pasynUser, &addr, 1);
    if (status != asynSuccess)
	return status;
    asynPrint (pasynUser, ASYN_TRACE_FLOW,
	       "%s samplerDriver:connect addr %d\n", pdrvPvt->portName, addr);
    if (addr >= 0)
      {
	  pasynManager->exceptionConnect (pasynUser);
	  return asynSuccess;
      }
    if (pdrvPvt->connected)
      {
	  asynPrint (pasynUser, ASYN_TRACE_ERROR,
		     "%s samplerDriver:connect port already connected\n",
		     pdrvPvt->portName);
	  return asynError;
      }
    pdrvPvt->connected = 1;
    pasynManager->exceptionConnect (pasynUser);
    return asynSuccess;
}
static asynStatus disconnect (void *pvt, asynUser * pasynUser)
{
    drvPvt_t *pdrvPvt = (drvPvt_t *) pvt;
    int addr;
    asynStatus status;
    status = getAddr (pdrvPvt, pasynUser, &addr, 1);
    if (status != asynSuccess)
	return status;
    asynPrint (pasynUser, ASYN_TRACE_FLOW,
	       "%s samplerDriver:disconnect addr %d\n", pdrvPvt->portName,
	       addr);
    if (addr >= 0)
      {
	  pasynManager->exceptionDisconnect (pasynUser);
	  return asynSuccess;
      }
    if (!pdrvPvt->connected)
      {
	  asynPrint (pasynUser, ASYN_TRACE_ERROR,
		     "%s samplerDriver:disconnect port not connected\n",
		     pdrvPvt->portName);
	  return asynError;
      }
    pdrvPvt->connected = 0;
    pasynManager->exceptionDisconnect (pasynUser);
    return asynSuccess;
}

#define PROCESSDATA(type) \
      {  \
	    if (psamplerPvt->onceMeasure.type > psamplerPvt->type.arraySize - 1)  \
		psamplerPvt->type.numGtLimit++; \
	    else \
	      psamplerPvt->type.array[psamplerPvt->onceMeasure.type]++; \
	    if (psamplerPvt->onceMeasure.type > psamplerPvt->type.max) \
		psamplerPvt->type.max = psamplerPvt->onceMeasure.type; \
	    if ((psamplerPvt->onceMeasure.type < psamplerPvt->type.min)  | (psamplerPvt->type.min==0) ) \
		psamplerPvt->type.min = psamplerPvt->onceMeasure.type; \
      }

#define PROCESS_MEDIAN(type) \
  { \
subTotal=0; \
for (j = 0; j <psamplerPvt->type.arraySize; j++) { \
subTotal += psamplerPvt->type.array[j]; \
if(subTotal > (psamplerPvt->totalSamples >> 1)) break; \
} \
psamplerPvt->type.median=j; \
}

#define CLEARDATA(type) \
    { \
    psamplerPvt->type.max=0; \
    psamplerPvt->type.min=0; \
    psamplerPvt->type.median=0; \
    psamplerPvt->type.numGtLimit = 0; \
    memset(psamplerPvt->type.array, 0, psamplerPvt->type.arraySize * sizeof( unsigned long)); \
}

static asynStatus processStartButton (drvPvt_t * pdrvPvt, ELLLIST * pclientList)
{
    int j;
    unsigned long subTotal = 0;
    interruptNode *pnode;
    asynInt32Interrupt *pinterrupt;

    for (j = 0; j < psamplerPvt->samples; j++)
      {
	  (*psamplerPvt->getSample) (psamplerPvt->samplerPvt,
				     &psamplerPvt->onceMeasure);
      PROCESSDATA (interruptLatency) PROCESSDATA (topToBottom)
	      PROCESSDATA (bottomToUserInterrupt)
	      PROCESSDATA (contextSwitch) PROCESSDATA (totalResponse)}
    psamplerPvt->totalSamples += psamplerPvt->samples;
    PROCESS_MEDIAN (interruptLatency);
    PROCESS_MEDIAN (topToBottom);
    PROCESS_MEDIAN (bottomToUserInterrupt);
    PROCESS_MEDIAN (contextSwitch);
    PROCESS_MEDIAN (totalResponse);

    pasynManager->interruptStart (pdrvPvt->asynInt32Pvt, &pclientList);
    pnode = (interruptNode *) ellFirst (pclientList);
    while (pnode)
      {
	  pinterrupt = pnode->drvPvt;
	  if (pinterrupt->addr == TOTAL_SAMPLES)
	    {
		pinterrupt->callback (pinterrupt->userPvt,
				      pinterrupt->pasynUser,
				      psamplerPvt->totalSamples);
	    }
	  pnode = (interruptNode *) ellNext (&pnode->node);
      }
    pasynManager->interruptEnd (pdrvPvt->asynInt32Pvt);
    return asynSuccess;
}
static asynStatus int32Write (void *pvt, asynUser * pasynUser, epicsInt32 value)
{
    drvPvt_t *pdrvPvt = (drvPvt_t *) pvt;
    drvUserData *pdrvUserData;
    int addr;
    asynStatus status;
    ELLLIST *pclientList;
    interruptNode *pnode;
    asynInt32Interrupt *pinterrupt;
    status = getAddr (pdrvPvt, pasynUser, &addr, 0);
    if (status != asynSuccess)
	return status;
    asynPrint (pasynUser, ASYN_TRACE_FLOW,
	       "%s samplerDriver:writeInt32 value %d\n", pdrvPvt->portName,
	       value);
    if (!pdrvPvt->connected)
      {
	  asynPrint (pasynUser, ASYN_TRACE_ERROR,
		     "%s samplerDriver:write not connected\n",
		     pdrvPvt->portName);
	  epicsSnprintf (pasynUser->errorMessage, pasynUser->errorMessageSize,
			 "%s samplerDriver:write not connected\n",
			 pdrvPvt->portName);
	  return asynError;
      }
    epicsMutexMustLock (psamplerPvt->lock);
    pdrvUserData = (drvUserData *) pasynUser->drvUser;

    switch (pdrvUserData->type)
      {
      case START:
	  processStartButton (pdrvPvt, pclientList);
	  break;
      case CLEAR:
	  psamplerPvt->totalSamples = 0;
	  CLEARDATA (interruptLatency);
	  CLEARDATA (topToBottom);
	  CLEARDATA (bottomToUserInterrupt);
	  CLEARDATA (contextSwitch);
	  CLEARDATA (totalResponse);

	  pasynManager->interruptStart (pdrvPvt->asynInt32Pvt, &pclientList);
	  pnode = (interruptNode *) ellFirst (pclientList);
	  while (pnode)
	    {
		pinterrupt = pnode->drvPvt;
		if (pinterrupt->addr == TOTAL_SAMPLES)
		  {
		      pinterrupt->callback (pinterrupt->userPvt,
					    pinterrupt->pasynUser,
					    psamplerPvt->totalSamples);
		  }
		pnode = (interruptNode *) ellNext (&pnode->node);
	    }
	  pasynManager->interruptEnd (pdrvPvt->asynInt32Pvt);
	  break;
      case SAMPLES:
	  psamplerPvt->samples = value;
	  break;
      case CUTOFF_H:
	  psamplerPvt->interruptLatency.cutoffH = value;
	  break;

      case CUTOFF_V:
	  psamplerPvt->interruptLatency.cutoffV = value;
	  break;
      case CUTOFF_H1:
	  psamplerPvt->contextSwitch.cutoffH = value;
	  break;
      case CUTOFF_V1:
	  psamplerPvt->contextSwitch.cutoffV = value;
	  break;
      case CUTOFF_H_TOP_TO_BH:
	  psamplerPvt->topToBottom.cutoffH = value;
	  break;
      case CUTOFF_V_TOP_TO_BH:
	  psamplerPvt->topToBottom.cutoffV = value;
	  break;
      case CUTOFF_H_BH_TO_USER:
	  psamplerPvt->bottomToUserInterrupt.cutoffH = value;
	  break;
      case CUTOFF_V_BH_TO_USER:
	  psamplerPvt->bottomToUserInterrupt.cutoffV = value;
	  break;
      case CUTOFF_H_TOTAL_RESPONSE:
	  psamplerPvt->totalResponse.cutoffH = value;
	  break;
      case CUTOFF_V_TOTAL_RESPONSE:
	  psamplerPvt->totalResponse.cutoffV = value;
	  break;
      default:
	  break;
      }
    epicsMutexUnlock (psamplerPvt->lock);
    asynPrint (pasynUser, ASYN_TRACEIO_DRIVER,
	       "%s addr %d write %d\n", pdrvPvt->portName, addr, value);
    return asynSuccess;
}
static asynStatus int32Read (void *pvt, asynUser * pasynUser,
			     epicsInt32 * value)
{
    drvPvt_t *pdrvPvt = (drvPvt_t *) pvt;
    drvUserData *pdrvUserData;
    int addr;
    asynStatus status;
    status = getAddr (pdrvPvt, pasynUser, &addr, 0);
    if (status != asynSuccess)
	return status;
    asynPrint (pasynUser, ASYN_TRACE_FLOW,
	       "%s samplerDriver:readInt32 value %d\n", pdrvPvt->portName,
	       value);
    if (!pdrvPvt->connected)
      {
	  asynPrint (pasynUser, ASYN_TRACE_ERROR,
		     "%s samplerDriver:read  not connected\n",
		     pdrvPvt->portName);
	  epicsSnprintf (pasynUser->errorMessage, pasynUser->errorMessageSize,
			 "%s samplerDriver:read not connected\n",
			 pdrvPvt->portName);
	  return asynError;
      }

    epicsMutexMustLock (psamplerPvt->lock);
    pdrvUserData = (drvUserData *) pasynUser->drvUser;
    switch (pdrvUserData->type)
      {
      case MAX_INTERRUPT_LATENCY:
	  *value = psamplerPvt->interruptLatency.max;
	  break;
      case MAX_CONTEXT_SWITCH:
	  *value = psamplerPvt->contextSwitch.max;
	  break;
      case MAX_TOP_TO_BH:
	  *value = psamplerPvt->topToBottom.max;
	  break;
      case MAX_BH_TO_USER:
	  *value = psamplerPvt->bottomToUserInterrupt.max;
	  break;
      case MAX_TOTAL_RESPONSE:
	  *value = psamplerPvt->totalResponse.max;
	  break;

      case NUM_GTM_INTERRUPT_LATENCY:
	  *value = psamplerPvt->interruptLatency.numGtLimit;
	  break;
      case NUM_GTM_CONTEXT_SWITCH:
	  *value = psamplerPvt->contextSwitch.numGtLimit;
	  break;
      case NUM_GTM_TOTAL_RESPONSE:
	  *value = psamplerPvt->totalResponse.numGtLimit;
	  break;
      case NUM_GTM_TOP_TO_BH:
	  *value = psamplerPvt->topToBottom.numGtLimit;
	  break;
      case NUM_GTM_BH_TO_USER:
	  *value = psamplerPvt->bottomToUserInterrupt.numGtLimit;
	  break;

      case MIN_INTERRUPT_LATENCY:
	  *value = psamplerPvt->interruptLatency.min;
	  break;
      case MIN_CONTEXT_SWITCH:
	  *value = psamplerPvt->contextSwitch.min;
	  break;
      case MIN_TOP_TO_BH:
	  *value = psamplerPvt->topToBottom.min;
	  break;
      case MIN_BH_TO_USER:
	  *value = psamplerPvt->bottomToUserInterrupt.min;
	  break;
      case MIN_TOTAL_RESPONSE:
	  *value = psamplerPvt->totalResponse.min;
	  break;

      case MEDIAN_INTERRUPT_LATENCY:
	  *value = psamplerPvt->interruptLatency.median;
	  break;
      case MEDIAN_CONTEXT_SWITCH:
	  *value = psamplerPvt->contextSwitch.median;
	  break;
      case MEDIAN_TOP_TO_BH:
	  *value = psamplerPvt->topToBottom.median;
	  break;
      case MEDIAN_BH_TO_USER:
	  *value = psamplerPvt->bottomToUserInterrupt.median;
	  break;
      case MEDIAN_TOTAL_RESPONSE:
	  *value = psamplerPvt->totalResponse.median;
	  break;

      case SAMPLES:
	  *value = psamplerPvt->samples;
	  break;

      case CUTOFF_H:
	  *value = psamplerPvt->interruptLatency.cutoffH;
	  break;

      case CUTOFF_V:
	  *value = psamplerPvt->interruptLatency.cutoffV;
	  break;
      case CUTOFF_H1:
	  *value = psamplerPvt->contextSwitch.cutoffH;
	  break;
      case CUTOFF_V1:
	  *value = psamplerPvt->contextSwitch.cutoffV;
	  break;
      case CUTOFF_H_TOP_TO_BH:
	  *value = psamplerPvt->topToBottom.cutoffH;
	  break;
      case CUTOFF_V_TOP_TO_BH:
	  *value = psamplerPvt->topToBottom.cutoffV;
	  break;
      case CUTOFF_H_BH_TO_USER:
	  *value = psamplerPvt->bottomToUserInterrupt.cutoffH;
	  break;
      case CUTOFF_V_BH_TO_USER:
	  *value = psamplerPvt->bottomToUserInterrupt.cutoffV;
	  break;
      case CUTOFF_H_TOTAL_RESPONSE:
	  *value = psamplerPvt->totalResponse.cutoffH;
	  break;
      case CUTOFF_V_TOTAL_RESPONSE:
	  *value = psamplerPvt->totalResponse.cutoffV;
	  break;
      default:
	  break;
      }
    epicsMutexUnlock (psamplerPvt->lock);
    asynPrint (pasynUser, ASYN_TRACEIO_DRIVER,
	       "%s read %d\n", pdrvPvt->portName, *value);
    return asynSuccess;
}

#define READ_ARRAY(type) \
{ \
        nremaining = psamplerPvt->type.arraySize; \
	pfrom = psamplerPvt->type.array; \
	while (nremaining > 0 && nout < nelements ) { \
	  if((*pfrom>=psamplerPvt->type.cutoffV && psamplerPvt->type.cutoffV !=0) || nout < psamplerPvt->type.cutoffH) *pto =0; \
	  else *pto=*pfrom; \
	  pto++; pfrom++; \
	  nremaining--; \
	  nout++; \
       } \
}

static asynStatus int32ArrayRead (void *pvt, asynUser * pasynUser,
				  epicsInt32 * value, size_t nelements,
				  size_t * nIn)
{
    drvPvt_t *pdrvPvt = (drvPvt_t *) pvt;
    drvUserData *pdrvUserData;
    epicsInt32 *pfrom, *pto;
    int nremaining;
    int nout = 0;
    int addr;
    asynStatus status;
    if (nIn)
	*nIn = 0;
    status = getAddr (pdrvPvt, pasynUser, &addr, 0);
    if (status != asynSuccess)
	return status;
    asynPrint (pasynUser, ASYN_TRACE_FLOW,
	       "%s %d uint32DigitalDriver:int32ArrayRead\n", pdrvPvt->portName,
	       addr);
    if (!pdrvPvt->connected)
      {
	  asynPrint (pasynUser, ASYN_TRACE_ERROR,
		     "%s uint32DigitalDriver:read  not connected\n",
		     pdrvPvt->portName);
	  epicsSnprintf (pasynUser->errorMessage, pasynUser->errorMessageSize,
			 "%s uint32DigitalDriver:read not connected\n",
			 pdrvPvt->portName);
	  return asynError;
      }
    epicsMutexMustLock (psamplerPvt->lock);
    pto = value;

    pdrvUserData = (drvUserData *) pasynUser->drvUser;
    if (pdrvUserData->type == INTERRUPT_LATENCY)
	READ_ARRAY (interruptLatency)
	else
    if (pdrvUserData->type == CONTEXT_SWITCH)
	READ_ARRAY (contextSwitch)
	else
    if (pdrvUserData->type == TOP_TO_BH)
	READ_ARRAY (topToBottom)
	else
    if (pdrvUserData->type == BH_TO_USER)
	READ_ARRAY (bottomToUserInterrupt)
	else
    if (pdrvUserData->type == TOTAL_RESPONSE)
	READ_ARRAY (totalResponse) if (nIn)
	    *nIn = nout;
    epicsMutexUnlock (psamplerPvt->lock);
    asynPrint (pasynUser, ASYN_TRACEIO_DRIVER,
	       "%s read %d\n", pdrvPvt->portName, value);
    return asynSuccess;
}

/* register samplerDriverInit*/
static const iocshArg samplerDriverInitArg0 = { "portName", iocshArgString };
static const iocshArg samplerDriverInitArg1 =
    { "arraySizeInterruptLatency", iocshArgInt };
static const iocshArg samplerDriverInitArg2 =
    { "arraySizeTopTobottom", iocshArgInt };
static const iocshArg samplerDriverInitArg3 =
    { "arraySizeBottomToUserInterrupt", iocshArgInt };
static const iocshArg samplerDriverInitArg4 =
    { "arraySizeContextSwitch", iocshArgInt };
static const iocshArg samplerDriverInitArg5 =
    { "arraySizeTotalResponse", iocshArgInt };

static const iocshArg *const samplerDriverInitArgs[6] =
    { &samplerDriverInitArg0, &samplerDriverInitArg1, &samplerDriverInitArg2,
    &samplerDriverInitArg3, &samplerDriverInitArg4, &samplerDriverInitArg5
};
static const iocshFuncDef samplerDriverInitFuncDef =
    { "samplerDriverInit", 6, samplerDriverInitArgs };
static void samplerDriverInitCallFunc (const iocshArgBuf * args)
{
    samplerDriverInit (args[0].sval, args[1].ival, args[2].ival, args[3].ival,
		       args[4].ival, args[5].ival);
}
static void samplerDriverRegister (void)
{
    static int firstTime = 1;
    if (firstTime)
      {
	  firstTime = 0;
	  iocshRegister (&samplerDriverInitFuncDef, samplerDriverInitCallFunc);
      }
}

epicsExportRegistrar (samplerDriverRegister);
