/* vxWorksSampler.c */
/* vxWorksSampler.c,v 1.1.1.1 2005/11/03 19:58:58 xusf Exp */
/***********************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of
* California, as Operator of Los Alamos National Laboratory, and
* Berliner Elektronenspeicherring-Gesellschaft m.b.H. (BESSY).
* asynDriver is distributed subject to a Software License Agreement
* found in file LICENSE that is included with this distribution.
***********************************************************************/
 
/*vxWorks dependent file
 * Author: Shifu Xu
 * Date:   03NOV2005
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <iocsh.h>
#include <epicsEvent.h>
#include <cantProceed.h>
#include <epicsExport.h>
#include <devLib.h>

#include "sampler.h"

typedef struct samplerOncePvt
{
    int vector;
    int level;
    double clockRateMhz;
    epicsEventId interruptEvent;
    unsigned long long tsUserInterrupt;
} samplerOncePvt;

static unsigned long long getTs (void)
{
    register long u;
    register long low;
    __asm__ __volatile__ ("mftb %1\n" "mftbu %0":"=r" (u), "=r" (low):);
    return (((unsigned long long) u) << 32) | ((unsigned long long) low);
}

static void samplerIntService (void *pvt)
{
    samplerOncePvt *psamplerOncePvt = (samplerOncePvt *) pvt;

    psamplerOncePvt->tsUserInterrupt = getTs ();
    epicsEventSignal (psamplerOncePvt->interruptEvent);
}
static int getSample (void *pvt, delayMeasure * pdelayMeasure)
{
    unsigned long long tsInterrupt;
    unsigned long long tsUserInterrupt;
    unsigned long long tsContextSwitch;
    int keys;

    samplerOncePvt *psamplerOncePvt = (samplerOncePvt *) pvt;
    keys = epicsInterruptLock ();
    tsInterrupt = getTs ();
    sysBusIntGen (psamplerOncePvt->level, psamplerOncePvt->vector);
    epicsInterruptUnlock (keys);
    epicsEventMustWait (psamplerOncePvt->interruptEvent);
    tsContextSwitch = getTs ();
    tsUserInterrupt = psamplerOncePvt->tsUserInterrupt;

    /* calculate */
    pdelayMeasure->interruptLatency =
	(unsigned long) ((unsigned long) ((tsUserInterrupt - tsInterrupt)) /
			 psamplerOncePvt->clockRateMhz);
    pdelayMeasure->contextSwitch =
	(unsigned long) ((unsigned long) ((tsContextSwitch - tsUserInterrupt)) /
			 psamplerOncePvt->clockRateMhz);
    pdelayMeasure->topToBottom = 0;
    pdelayMeasure->totalResponse =
	(unsigned long) ((unsigned long) ((tsContextSwitch - tsInterrupt)) /
			 psamplerOncePvt->clockRateMhz);
    return 0;
}

void platformSamplerInit (const char *name, int vector, int level,
			  double clockRateMhz)
{
    samplerOncePvt *psamplerOncePvt;

    psamplerOncePvt = callocMustSucceed (1, sizeof (samplerOncePvt), "name");
    psamplerOncePvt->level = level;
    psamplerOncePvt->vector = vector;
    psamplerOncePvt->clockRateMhz = clockRateMhz;
    psamplerOncePvt->interruptEvent = epicsEventCreate (epicsEventEmpty);
    psamplerOncePvt->tsUserInterrupt = 0;
    if (devConnectInterruptVME
	(vector, samplerIntService, (void *) psamplerOncePvt) != OK)
      {
	  epicsPrintf (" sampler intConnect failed");
	  return;
      }
    devEnableInterruptLevelVME (level);
    psamplerManager->registerSampler (name, getSample, psamplerOncePvt);
}

void platformSamplerQuit (void *pvt)
{
    samplerOncePvt *psamplerOncePvt = (samplerOncePvt *) pvt;
    psamplerManager->removeSampler (pvt);
    free (psamplerOncePvt);
}

static const iocshArg initArg0 = { "platformSamplerInit", iocshArgString };
static const iocshArg initArg1 = { "interrupt vector", iocshArgInt };
static const iocshArg initArg2 = { "interrupt level", iocshArgInt };
static const iocshArg initArg3 = { "clock rate in Mhz", iocshArgDouble };

static const iocshArg *const initArgs[4] =
    { &initArg0, &initArg1, &initArg2, &initArg3 };
static void initCallFunc (const iocshArgBuf * args)
{
    platformSamplerInit (args[0].sval, args[1].ival, args[2].ival,
			 args[3].dval);
}

static const iocshFuncDef initFuncDef = { "platformSamplerInit", 4, initArgs };

static void platformSamplerRegister (void)
{
    static int firstTime = 1;
    if (!firstTime)
	return;
    firstTime = 0;
    iocshRegister (&initFuncDef, initCallFunc);
}

epicsExportRegistrar (platformSamplerRegister);
